/*
 * This file is part of PolicyUpdater.
 *
 * Copyright (C) 2003, 2004, 2005 University of Western Sydney
 * by Vino Fernando Crescini <jcrescin@scm.uws.edu.au>
 *
 * PolicyUpdater is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * PolicyUpdater is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PolicyUpdater; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <new>

#include <getopt.h>

#include <vlad/vlad.h>
#include "parser.h"

void print_usage(FILE *a_stream);

int main(int argc, char *argv[])
{
  int retval;
  char arglist[5];
  int curr_opt;
  FILE *programin = NULL;
  FILE *operationin = NULL;
  kb *kbase = NULL;
  unsigned char mode = VLAD_MODE_GENERATE;
#ifdef VLAD_TIMER
  unsigned char timer = 0;
#endif

  /* compose arglist */
  strcpy(arglist, "vh");
#ifdef VLAD_SMODELS
  strcat(arglist, "e");
#endif
#ifdef VLAD_TIMER
  strcat(arglist, "t");
#endif

  opterr = 0;

  while ((curr_opt = getopt(argc, argv, arglist)) != -1) {
    switch(curr_opt) {
      case 'v' :
        fprintf(stdout, "vLad %s\n", VLAD_VERSION);
        fprintf(stdout, "Copyright (C) 2003, 2004, 2005 University of Western Sydney\n");
        fprintf(stdout, "by Vino Fernando Crescini  <jcrescin@scm.uws.edu.au>\n");
        return VLAD_OK;
      case 'h' :
        print_usage(stdout);
        return VLAD_OK;;
#ifdef VLAD_SMODELS
      case 'e' :
	mode = VLAD_MODE_EVALUATE;
	break;
#endif
#ifdef VLAD_TIMER
      case 't' :
        timer = 1;
        break;
#endif
      default :
        print_usage(stderr);
        return VLAD_FAILURE;
    }
  }

  /* get program file details */
  if (argv[optind] == NULL || !strcmp(argv[optind], "-")) {
    print_usage(stderr);
    return VLAD_FAILURE;
  }

  if ((programin = fopen(argv[optind], "r")) == NULL) {
    fprintf(stderr, "unable to open for reading: %s\n", argv[optind]);
    return VLAD_OPENFAILED;
  }

  /* now get operation file details */
  if (optind + 1 >= argc || argv[optind + 1] == NULL || !strcmp(argv[optind + 1], "-"))
    operationin = stdin;
  else if ((operationin = fopen(argv[optind + 1], "r")) == NULL) {
    fprintf(stderr, "unable to open for reading: %s\n", argv[optind + 1]);
    return VLAD_OPENFAILED;
  }

  /* create an instance of the kb and initialise it */
  if ((kbase = VLAD_NEW(kb)) == NULL)
    return VLAD_MALLOCFAILED;

  if ((retval = kbase->init()) != VLAD_OK) {
    fprintf(stderr, "cannot initialise kbase: %d\n", retval);
    return retval;
  }

  /* first initialise the parsers */
  if ((retval = program_init(programin, stdout, stderr, kbase)) != VLAD_OK)
    return retval;

#ifdef VLAD_TIMER
  retval = operation_init(operationin,
                          stdout,
                          stderr,
                          kbase,
                          mode,
                          timer);
#else
  retval = operation_init(operationin,
                          stdout,
                          stderr,
                          kbase,
                          mode);
#endif
  if (retval != VLAD_OK)
    return retval;

  /* then parse */
  if ((retval = program_parse()) != VLAD_OK)
    return retval;
  if ((retval = operation_parse()) != VLAD_OK)
    return retval;

  /* cleanup */
  fclose(programin);
  fclose(operationin);
  delete kbase;

  return VLAD_OK;
}

void print_usage(FILE *a_stream)
{
  if (a_stream == NULL)
    return;

  fprintf(a_stream, "usage: vlad [options] program-filename [operation-filename]\n");
  fprintf(a_stream, "  -v     display version number\n");
  fprintf(a_stream, "  -h     display this message\n");
#ifdef VLAD_SMODELS
  fprintf(a_stream, "  -e     if configured with --with-smodels, evaluate queries\n");
#endif
#ifdef VLAD_TIMER
  fprintf(a_stream, "  -t     if configured with --enable-timer, display computation time\n");
#endif
}
